@@ -1,5 +1,6 @@
 from __future__ import annotations
 
+import math
 import os
 import random
 import re
@@ -22,8 +23,8 @@ from requests.exceptions import InvalidSchema
 
 from isisdl.backend.crypt import get_credentials
 from isisdl.backend.status import StatusOptions, DownloadStatus, RequestHelperStatus
-from isisdl.settings import download_timeout, download_timeout_multiplier, download_static_sleep_time, num_tries_download, status_time, perc_diff_for_checksum, error_text, bandwidth_mavg_perc, \
-    extern_ignore, log_file_location, datetime_str, regex_is_isis_document
+from isisdl.settings import download_timeout, download_timeout_multiplier, download_static_sleep_time, num_tries_download, status_time, perc_diff_for_checksum, error_text, extern_ignore, \
+    log_file_location, datetime_str, regex_is_isis_document, token_queue_download_refresh_rate, download_chunk_size, download_progress_bar_resolution, bandwidth_download_files_mavg_perc
 from isisdl.settings import enable_multithread, discover_num_threads, is_windows, is_macos, is_testing, testing_bad_urls, url_finder, isis_ignore
 from isisdl.utils import User, path, sanitize_name, args, on_kill, database_helper, config, generate_error_message, logger, parse_google_drive_url, get_url_from_gdrive_confirmation, \
     DownloadThrottler, MediaType, HumanBytes, normalize_url
@@ -411,15 +412,18 @@ class MediaContainer:
         if isinstance(maybe_container, bool):
             return maybe_container
 
-        if maybe_container.checksum is not None:
-            return False
-
-        if actual_size == 0 or not (self.size * (1 - perc_diff_for_checksum) <= actual_size <= self.size * (1 + perc_diff_for_checksum)):
+        if actual_size == 0:
             return True
 
         if self.size == actual_size:
             return False
 
+        if maybe_container.checksum is None:
+            return True
+
+        if self.size * (1 - perc_diff_for_checksum) <= actual_size <= self.size * (1 + perc_diff_for_checksum):
+            return False
+
         return calculate_local_checksum(self.path) == maybe_container.checksum
 
     def hardlink(self, other: MediaContainer) -> None:
@@ -451,6 +455,29 @@ class MediaContainer:
 
         return sanitize_name(self._name, False)
 
+    def render_progress_bar(self) -> str:
+        if self.size in {0, -1}:
+            percent: float = 0.
+        elif self.current_size is None:
+            percent = 0.
+        else:
+            percent = self.current_size / self.size
+
+        # Sometimes this bug happens… I don't know why and I don't feel like debugging it.
+        if percent > 1:
+            percent = 1.
+
+        progress_chars = int(percent * download_progress_bar_resolution)
+        return "╶" + "█" * progress_chars + " " * (download_progress_bar_resolution - progress_chars) + "╴"
+
+    def render_status(self, course_pad: int = 0, hostname_pad: int = 0, stream: bool = False) -> str:
+        return \
+            f"{'Stream:  ' if stream else ''}{self.render_progress_bar()} " \
+            f"[ {HumanBytes.format_pad(self.current_size)} | {HumanBytes.format_pad(self.size)} ]" \
+            f" - {str(self.course):<{course_pad}}" \
+            f" - {urlparse(self.url).hostname:<{hostname_pad}}" \
+            f" - {self}"
+
     def __repr__(self) -> str:
         return self.__str__()
 
@@ -479,20 +506,27 @@ class MediaContainer:
     def stop(self) -> None:
         self._stop = True
 
-    def download(self, throttler: DownloadThrottler, session: SessionWithKey, is_stream: bool = False) -> None:
+    def download(self, throttler: DownloadThrottler, session: SessionWithKey, is_stream: bool = False) -> bool:
+        """
+        The bool return value indicates if any downloading took place. Used for the bandwidth calculations.
+        """
+
         if self._stop or self.media_type == MediaType.corrupted:
             self._done = True
-            return
+            return False
 
         if self.current_size is not None:
-            return
+            if is_testing:
+                assert self._done
+
+            return False
 
         self.current_size = 0
 
         if not self.should_download and self.url != 'https://www.eecs.tu-berlin.de/fileadmin/f4/fkIVdokumente/studium/Plagiate/Merkblatt_Plagiate_Fak.IV_05-2020.pdf':
             self.current_size = self.size
             self._done = True
-            return
+            return False
 
         if is_stream:
             throttler.start_stream(self.path)
@@ -500,10 +534,14 @@ class MediaContainer:
         download = session.get_(self.download_url, params={"token": session.token}, stream=True)
 
         if download is None or not download.ok:
+            if download is not None:
+                download.close()
+
             self.size = 0
             self.current_size = None
             self.media_type = MediaType.corrupted
             self._done = True
+
             if self.media_type != MediaType.video:
                 # The video server is sometimes unreliable but it _should_ always work. So don't add these url's
                 database_helper.add_bad_url(self.url)
@@ -513,7 +551,7 @@ class MediaContainer:
                     link.hardlink(self)
 
             self.dump()
-            return
+            return False
 
         # We copy in chunks so the download rate can be limited. This could also be done with `shutil.copyfileobj(…)`
         with self.path.open("wb") as f:
@@ -549,9 +587,8 @@ class MediaContainer:
         download.close()
 
         # Only register the file after successfully downloading it.
-        if self.media_type != MediaType.corrupted:
-            if is_testing:
-                assert self.size * (1 - perc_diff_for_checksum) <= self.path.stat().st_size <= self.size * (1 + perc_diff_for_checksum), self.path
+        if is_testing and self.media_type != MediaType.corrupted:
+            assert self.size * (1 - perc_diff_for_checksum) <= self.path.stat().st_size <= self.size * (1 + perc_diff_for_checksum), self.path
 
         self.size = self.path.stat().st_size
         self.checksum = calculate_local_checksum(self.path)
@@ -569,6 +606,8 @@ class MediaContainer:
         self._newly_downloaded = True
         self._done = True
 
+        return True
+
 
 class Course:
     displayname: str
@@ -871,30 +910,31 @@ class RequestHelper:
             all_content = []
             _assignments = self.post_REST("mod_assign_get_assignments", use_timeout=False)
             if _assignments is None:
-                assignments = cast(Dict[str, Any], _assignments)
+                return []
 
-                allowed_ids = {item.course_id for item in self.courses}
-                for _course in assignments["courses"]:
-                    if _course["id"] in allowed_ids:
-                        for assignment in _course["assignments"]:
-                            if "introattachments" not in assignment:
-                                continue
+            assignments: Dict[str, Any] = _assignments
 
-                            for file in assignment["introattachments"]:
-                                file["filepath"] = assignment["name"]
-                                all_content.append(PreMediaContainer(
+            allowed_ids = {item.course_id for item in self.courses}
+            for _course in assignments["courses"]:
+                if _course["id"] in allowed_ids:
+                    for assignment in _course["assignments"]:
+                        if "introattachments" not in assignment:
+                            continue
+
+                        for file in assignment["introattachments"]:
+                            file["filepath"] = assignment["name"]
+                            all_content.append(
+                                PreMediaContainer(
                                     file["fileurl"], RequestHelper.course_id_mapping[_course["id"]], MediaType.document,
                                     file["filename"], file["filepath"], file["filesize"], file["timemodified"])
-                                )
+                            )
 
-                return all_content
+            return all_content
 
         except Exception as ex:
             with self._lock:
                 generate_error_message(ex)
 
-        return []
-
     def _download_videos(self, _: Any) -> List[PreMediaContainer]:
         try:
             if config.download_videos is False:
@@ -1024,7 +1064,7 @@ def check_for_conflicts_in_files(files: List[MediaContainer]) -> List[MediaConta
 
 class Downloader(Thread):
     q: Queue[MediaContainer]
-    ret_q: Queue[int]
+    ret_q: Queue[Tuple[int, bool]]
     thread_id: int
     status: DownloadStatus
     throttler: DownloadThrottler
@@ -1033,7 +1073,7 @@ class Downloader(Thread):
     is_downloading: bool = False
     _lock = Lock()
 
-    def __init__(self, q: Queue[MediaContainer], ret_q: Queue[int], thread_id: int, status: DownloadStatus, throttler: DownloadThrottler, session: SessionWithKey):
+    def __init__(self, q: Queue[MediaContainer], ret_q: Queue[Tuple[int, bool]], thread_id: int, status: DownloadStatus, throttler: DownloadThrottler, session: SessionWithKey):
         self.q = q
         self.ret_q = ret_q
         self.thread_id = thread_id
@@ -1051,41 +1091,16 @@ class Downloader(Thread):
                 self.is_downloading = True
                 self.status.add_container(self.thread_id, file)
 
-                file.download(self.throttler, self.session)
+                ret = file.download(self.throttler, self.session)
                 self.status.done(self.thread_id, file)
                 self.is_downloading = False
-                self.ret_q.put(self.thread_id)
+                self.ret_q.put((self.thread_id, ret))
 
         except Exception as ex:
             with self._lock:
                 generate_error_message(ex)
 
 
-class BandwidthWatcher(Thread):
-    throttler: DownloadThrottler
-    num_threads: int
-    bandwidths: Dict[int, Tuple[int, float]]
-
-    def __init__(self, throttler: DownloadThrottler, num_threads: int, bandwidths: Dict[int, Tuple[int, float]]) -> None:
-        self.throttler = throttler
-        self.num_threads = num_threads
-        self.bandwidths = bandwidths
-
-        super().__init__(daemon=True)
-        self.start()
-
-    def run(self) -> None:
-        while True:
-            num_samples, old_bw = self.bandwidths[self.num_threads]
-            if num_samples:
-                new_bw = old_bw * (1 - bandwidth_mavg_perc) + self.throttler.bandwidth_used * bandwidth_mavg_perc
-                self.bandwidths[self.num_threads] = (num_samples + 1, new_bw)
-            else:
-                self.bandwidths[self.num_threads] = (1, self.throttler.bandwidth_used)
-
-            pass
-
-
 class CourseDownloader:
     containers: Dict[MediaType, List[MediaContainer]] = {}
     _did_message: bool = False
@@ -1241,7 +1256,7 @@ Newly discovered files:
         # TODO: Dynamic calculation of num threads such that optimal Internet Usage is achieved
         exception_lock = Lock()
 
-        def download(file: MediaContainer) -> None:
+        def download(file: MediaContainer) -> bool:
             if enable_multithread:
                 thread_id = int(current_thread().name.split("T_")[-1])
             else:
@@ -1249,13 +1264,15 @@ Newly discovered files:
 
             status.add_container(thread_id, file)
             try:
-                file.download(throttler, session)
+                exit_ = file.download(throttler, session)
                 status.done(thread_id, file)
+                return exit_
 
             except Exception as ex:
                 with exception_lock:
                     generate_error_message(ex)
 
+        # TODO: Refactor this into a `sorted` expression with file.should_download as key.
         first_files: List[MediaContainer] = []
         second_files: List[MediaContainer] = []
 
@@ -1273,43 +1290,100 @@ Newly discovered files:
             for file in first_files + second_files:
                 download(file)
 
-        # files_to_download = files[MediaType.document] + files[MediaType.extern] + files[MediaType.video]
-        # files_to_download = [file for file in files_to_download if file.should_download]
-        #
-        # bandwidths: Dict[int, Tuple[int, float]] = {i: (0, 0.0) for i in range(args.max_num_threads)}
-        # ret_q: Queue[int] = Queue()
-        #
-        # if enable_multithread:
-        #     threads = [Downloader(Queue(), ret_q, i, status, throttler, session) for i in range(args.max_num_threads)]
-        #
-        # else:
-        #     threads = [Downloader(Queue(), ret_q, 0, status, throttler, session)]
-        #
-        # max_downloading = args.max_num_threads // 2
-        # i = -1
-        # num_iter = 2 / status_time
-        #
-        # while files_to_download:
-        #     i += 1
-        #
-        #     num_dl_threads = sum(thread.is_downloading for thread in threads)
-        #
-        #
-        #     if i == num_iter:
-        #         i = 0
-        #
-        #     if num_dl_threads >= max_downloading:
-        #         time.sleep(status_time)
-        #         continue
-        #
-        #     for thread in threads:
-        #         if num_dl_threads >= max_downloading:
-        #             break
-        #
-        #         if not thread.is_downloading:
-        #             thread.q.put(files_to_download.pop(0))
-        #             time.sleep(status_time)
-        #             num_dl_threads += 1
+    def _download_files(self, files: Dict[MediaType, List[MediaContainer]], throttler: DownloadThrottler, session: SessionWithKey, status: DownloadStatus) -> None:
+
+        # The threads only have to be re-filled once there is an item in ret_q.
+        # → The entire process could be blocking instead of active waiting
+        # → One could always store the bandwidth from before and evaluate on waking if there was a speedup
+
+        # We have to be explorative in + and - directions:
+        #   If we are currently at the maximum number of threads recorded, then sometimes just add a thread.
+
+        # Too many threads don't hurt: They just Make stopping the download slower.
+
+        # Maybe add in two phases: First download all small files → (Everything below 30M?), then the big one.
+        # Use the same algorithm but with the maximum number of threads doubled.
+
+        # TODO: Add extern + corrupted + documents
+        files_to_download = files[MediaType.video]
+        files_to_download.sort(key=lambda it: it.should_download)
+
+        ret_q: Queue[Tuple[int, bool]] = Queue()
+
+        if enable_multithread:
+            threads = [Downloader(Queue(), ret_q, i, status, throttler, session) for i in range(args.max_num_threads)]
+
+        else:
+            threads = [Downloader(Queue(), ret_q, 0, status, throttler, session)]
+
+        i = 0
+        for i in range(args.max_num_threads // 2):
+            threads[i].q.put(files_to_download.pop(0))
+
+        num_threads_downloading = i + 1
+
+        def eval_spawn_next_thread() -> Optional[bool]:
+            # TODO
+            # 2. Predict in which direction to 'jump'
+
+            num_times_counted
+            bandwidths
+
+            weights: Dict[Optional[bool], float] = {False: 0.0, True: 0.0, None: 0.0}
+
+            # Score metric consisting of
+            #   - number of times already sampled
+            #   - Expected utility → Fewer threads = better
+            if num_threads_downloading + 1 in bandwidths:
+                weights[True] += -2 * math.log(num_times_counted[num_threads_downloading + 1]) + 2
+
+            else:
+                weights[True] += 3
+
+            if num_threads_downloading - 1 in bandwidths:
+                weights[False] += -2 * math.log(num_times_counted[num_threads_downloading - 1]) + 2
+
+            else:
+                weights[False] += 3
+
+            choice = random.choices(list(weights.keys()), list(weights.values()))
+            return choice[0]
+
+        bandwidths: Dict[int, float] = {}
+        num_times_counted: Dict[int, int] = {}
+        time_last_measurement = time.perf_counter()
+
+        while files_to_download:
+            thread_id_woken_up, was_successful = ret_q.get()
+
+            if was_successful:
+                # Measure the bandwidth
+                time_taken = min(time.perf_counter() - time_last_measurement, token_queue_download_refresh_rate)
+                timestamps = [item for item in throttler.timestamps if time.perf_counter() - item < time_taken]
+
+                bandwidth_used = float(len(timestamps) * download_chunk_size / token_queue_download_refresh_rate)
+
+                if num_threads_downloading not in bandwidths:
+                    bandwidths[num_threads_downloading] = bandwidth_used
+                    num_times_counted[num_threads_downloading] = len(timestamps)
+                else:
+                    bandwidths[num_threads_downloading] = bandwidths[num_threads_downloading] * (1 - bandwidth_download_files_mavg_perc) + bandwidth_used * bandwidth_download_files_mavg_perc
+                    num_times_counted[num_threads_downloading] += len(timestamps)
+
+            spawn = eval_spawn_next_thread()
+
+            if spawn is True and num_threads_downloading < args.max_num_threads:
+                threads[thread_id_woken_up].q.put(files_to_download.pop(0))
+                threads[num_threads_downloading].q.put(files_to_download.pop(0))
+                num_threads_downloading += 1
+
+            elif spawn is False and num_threads_downloading > 1:
+                num_threads_downloading -= 1
+
+            else:
+                threads[thread_id_woken_up].q.put(files_to_download.pop(0))
+
+            time_last_measurement = time.perf_counter()
 
     @staticmethod
     @on_kill(2)
