@@ -39,7 +39,8 @@ from requests import Session
 
 from isisdl import settings
 from isisdl.backend.database_helper import DatabaseHelper
-from isisdl.settings import download_chunk_size, token_queue_download_refresh_rate, forbidden_chars, replace_dot_at_end_of_dir_name, force_filesystem, has_ffmpeg, fstype, log_file_location
+from isisdl.settings import download_chunk_size, token_queue_download_refresh_rate, forbidden_chars, replace_dot_at_end_of_dir_name, force_filesystem, has_ffmpeg, fstype, log_file_location, \
+    source_code_location
 from isisdl.settings import working_dir_location, is_windows, checksum_algorithm, checksum_num_bytes, example_config_file_location, config_dir_location, database_file_location, status_time, \
     discover_num_threads, status_progress_bar_resolution, download_progress_bar_resolution, config_file_location, is_first_time, is_autorun, parse_config_file, lock_file_location, \
     enable_lock, error_directory_location, systemd_dir_location, master_password, is_testing, systemd_timer_file_location, systemd_service_file_location, export_config_file_location, \
@@ -388,6 +389,7 @@ def startup() -> None:
         os.makedirs(path(config_dir_location), exist_ok=True)
         os.makedirs(systemd_dir_location, exist_ok=True)
 
+        # Install config file
         default_config_str = generate_default_config_str()
 
         with open(example_config_file_location, "w") as f:
@@ -397,6 +399,55 @@ def startup() -> None:
             with open(config_file_location, "w") as f:
                 f.write(f"# You probably want to start by copying {example_config_file_location} and adapting it.\n")
 
+    if not is_windows and not is_static:
+        # Install completions if they don't exist already
+        user_dir = os.path.expanduser("~")
+
+        if shutil.which("zsh") is not None:
+            default_path = f"{user_dir}/.local/share/zsh/site-functions"
+            whitelist_paths = [
+                f"{user_dir}/.oh-my-zsh/completions",
+                f"{user_dir}/.oh-my-zsh/custom/plugins/zsh-completions/src",
+            ]
+
+            # Since the session is interactive it might contain stuff like neofetch etc.
+            rand_string = "".join(random.choice(string.ascii_letters) for _ in range(32))
+            fpath = subprocess.check_output(["zsh", "-c", "-i", f"echo {rand_string}; echo $FPATH"]) \
+                .decode().split(rand_string)[-1].strip().split(":")
+
+            # 1. Find install path (default .local/share/zsh/site-functions or .local/share/isisdl/)
+            # 2. copy _isisdl there
+
+            # 3. Export this path to $FPATH via .zshrc (make this transparent to user)
+            # 4. autoload [path]/_isisdl
+
+            for final_path in whitelist_paths:
+                for p in fpath:
+                    if final_path == p and not (os.path.exists(final_path) and not os.access(final_path, os.W_OK | os.X_OK)):
+                        break
+
+            else:
+                final_path = default_path
+
+            # Here the file is always copied to the destination dir.
+            # This is fine since the Filesystem will (hopefully) realize that the files are the same and thus ignore it.
+            os.makedirs(final_path, exist_ok=True)
+
+            completion_source = source_code_location.joinpath("resources", "completions", "zsh", "_isisdl")
+            if completion_source.exists():
+                shutil.copy(completion_source, final_path)
+
+            # TODO: Also tag this file as autoloadable
+
+        if shutil.which("fish") is not None:
+            # TODO
+            pass
+
+        if shutil.which("bash") is not None:
+            final_path = f"{user_dir}/.local/share/bash-completion/completions"
+            os.makedirs(final_path, exist_ok=True)
+            shutil.copy(source_code_location.joinpath("resources", "completions", "zsh", "_isisdl"), final_path)
+
 
 def clear() -> None:
     if is_windows:
@@ -602,9 +653,6 @@ def check_github_for_version() -> Optional[Union[LegacyVersion, Version]]:
 
 
 def print_changelog_for_version(new_version: Union[LegacyVersion, Version]) -> None:
-    def version_check(v: str) -> bool:
-        return version.parse(__version__) < version.parse(v) <= new_version
-
     current_version = version.parse(__version__)
     assert not isinstance(current_version, LegacyVersion)
     assert not isinstance(new_version, LegacyVersion)
@@ -612,7 +660,9 @@ def print_changelog_for_version(new_version: Union[LegacyVersion, Version]) -> N
     for i in range(current_version.micro + 1, new_version.micro + 1):
         con = requests.get(f"https://raw.githubusercontent.com/Emily3403/isisdl/main/src/isisdl/resources/changelog/"
                            f"{new_version.major}.{new_version.minor}/{new_version.major}.{new_version.minor}.{new_version.micro}.md")
+        print("\n-----\n")
         print(con.text)
+        print("\n-/---")
 
 
 def install_latest_version() -> None:
@@ -807,6 +857,10 @@ def sanitize_name(name: str, is_dir: bool) -> str:
 
     final_str = "".join(item for item in final if item not in forbidden_chars)
 
+    # If the final name is ".." and it is interpreted as a directory this leads to a directory traversal vulnerability.
+    if is_dir and final_str == "..":
+        final_str = "DotDot"
+
     # All supported filesystems have a limit of 255 bytes for the filename. Enforce it.
     return final_str.encode()[:255].decode()
 
@@ -1360,7 +1414,7 @@ class DownloadThrottler(Thread):
         return int((self.download_rate * 1024 ** 2) // download_chunk_size * self.refresh_rate) or 1
 
 
-# TODO: Add link type and store that in the database + reference to checksum?
+# TODO: Add link type and store that in the database + reference to checksum?<
 class MediaType(enum.Enum):
     document = 1
     extern = 2
